package imie;

/**
 * @author Jacek Piechota
 * email opta@opta.com.pl
 * @version 1.0
 * @since 2.19.18.1.12 5 Eb 5 Pax
 * Klasa opisująca datę uniwersalną, zarówno juliańską jak i gregoriańską.
 * Rodzaj daty jest rozpoznawany po dacie. Daty zapisywane są "astronomicznie"
 * tzn. mogą być ujemne (zakłada się, że istnieje rok 0).
 * Start Kalendarza Juliańskiego był -4712,1,1.5 czyli 1 stycznia roku 4713 
 * p.n.e, i jest nazwany Julian Day 0 (JD0). 
 * Liczba 1.5 oznacza, że data liczona jest od godziny 12.00 w południe (czyli
 * po przecinku podana jest ułamkowa frakcja dnia).  Czas tutaj podany jest
 * UT (Universal Time). Po 
 * dacie 4 Oct 1582 następuje od razu data 15 X 1582 i jest to początek 
 * kalendarza gregoriańskiego. Różnica między kalendarzem juliańskim i
 * gregoriańskim jest w sposobie uznawania lat za przestępne. Lata przestępne
 * w kalendarzu gregoriańskim są liczone tak samo jak w juliańskim (czyli
 * każdy rok, który dzieli się przez 4) z wyjątkiem dat mających 00 na końcu. 
 * Lata, które dzielą się przez 100, ale nie przez 400, są zwykłymi latami. 
 * Lat podzielne przez 100 i przez 400 są latami przestępnymi. Czyli rok 1700, 
 * 1800, 1900 były  zwykłe, a lata 1600 i 2000 były przestępne.
 * Kalendarz Gregoriański był w różnych krajach wprowadzony w różnych 
 * terminach. W Polsce był wprowadzony 15 X 1582. W krajach, w których 
 * kalendarz gregoriański wprowadzono w innym terminie konieczna jest zmiana 
 * niektórych algorytmów w tej klasie. 
 */
public class YMD implements Comparable<YMD>{
	/**
	 * Tablica zawierająca liczbę dni miesiąca w roku normalnym
	 */
	private static final int[] LEN = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31,
		30, 31};
	/**
	 * Tablica zawierająca liczbę dni  miesiąca w roku przestępnym;
	 */
	private static final int[] LEAP_LEN = {31, 29, 31, 30, 31, 30, 31, 31, 30,
		31, 30, 31};
	/**
	 * Rok. Może być liczbą ujemną. Zawsze 4 cyfry.
	 */
	private int yy;
	/**
	 * Miesiąc. <1-12>. Bez zer początkowych.
	 */
	private int mm;
	/**
	 * Dzień miesiąca. <1,31>. Zawsze bez zer początkowych. Część po przecinku
	 * oznacza ułamek dziesiętny dnia.
	 */
	private double dd;

	//----------------------- constructors -----------------------------------
	/**
	 * Konstruktor 
	 * @param yy - rok. Może być liczbą ujemną. Zawsze 4 cyfry.
	 * @param mm - miesiac. <1-12>. Bez zer początkowych.
	 * @param dd - dzień miesiąca. <1,31>. Zawsze bez zer początkowych. 
	 * Część po przecinku oznacza ułamek dziesiętny dnia.
	 */
	public YMD(int yy, int mm, double dd){
		if(isValidDate(yy, mm, dd)){
			this.yy = yy;
			this.mm = mm;
			this.dd = dd;
		}
		else{
			throw new IllegalArgumentException("nieprawidłowa data");
		}
	}

	//----------------------------- methods ---------------------------------
	/**
	 * Dodaje wskazaną liczbę dni do tej daty
	 * @param daysAdded - liczba dni, które należy dodać. Liczba dni dodanych
	 * nie może być ujemna
	 */
	public void addDays(double daysAdded) {
		if(daysAdded < 0){
			throw new IllegalArgumentException("Liczba dni dodanych nie może "
					+ "być ujemna");
		}
		double jd = dateToJulianDay(yy, mm, dd);
		jd += daysAdded;
		YMD result = julianDayToDate(jd);
		yy = result.getYy();
		mm = result.getMm();
		dd = roundToDecimal(result.getDd(), 4);
	}

	/**
	 * Odejmuje wskazaną liczbę dni od tej daty. 
	 * @param daysSubstracted - liczba dni, które należy odjąć. Liczba dni 
	 * odejmowanych nie może być ujemna.
	 */
	public void substractDays(double daysSubstracted) {
		if(daysSubstracted < 0){
			throw new IllegalArgumentException("Liczba dni odejmowanych nie "
					+ "może być ujemna");
		}
		double jd = dateToJulianDay(yy, mm, dd);
		jd -= daysSubstracted;
		YMD result = julianDayToDate(jd);
		if(isValidDate(result)){
			yy = result.getYy();
			mm = result.getMm();
			dd = roundToDecimal(result.getDd(), 4);
		}
		else{
			throw new IllegalArgumentException("data poza zakresem kalendarza");
		}
	}

	/** 
	 * Oblicza liczbę dni między tą datą, a podaną datą
	 * @param y - rok
	 * @param m - miesiąc
	 * @param d - dzień
	 * @return - zwraca liczbę dni między tą datą, a podaną datą.
	 */
	public double daysElapsed(int y, int m, double d) {
		double jd1 = dateToJulianDay(yy, mm, dd);
		double jd2 = dateToJulianDay(y, m, d);
		return Math.abs(jd2 - jd1);
	}

	/** 
	 * Oblicza liczbę dni między tą datą, a podaną datą
	 * @param date - data
	 * @return - zwraca liczbę dni między tą datą, a podaną datą.
	 */
	public double daysElapsed(YMD date) {
		double jd1 = dateToJulianDay(yy, mm, dd);
		double jd2 = dateToJulianDay(date);
		return Math.abs(jd2 - jd1);
	}

	//------------------------ setters && getters ----------------------------
	/**
	 * Pobiera rok
	 * @return - zwraca rok
	 */
	public int getYy() {
		return yy;
	}

	/**
	 * Pobiera miesiąc
	 * @return - zwraca miesiąć
	 */
	public int getMm() {
		return mm;
	}

	/**
	 * Pobiera dzień
	 * @return - zwraca dzień
	 */
	public double getDd() {
		return dd;
	}

	public void setYy(int yy) {
		if(isValidDate(yy, this.mm, this.dd)){
			this.yy = yy;
		}
	}

	public void setMm(int mm) {
		if(isValidDate(this.yy, mm, this.dd)){
			this.mm = mm;
		}
	}

	public void setDd(double dd) {
		if(isValidDate(this.yy, this.mm, dd)){
			this.dd = dd;
		}
	}

	//----------------------------- overrides ---------------------------------
	@Override
	public boolean equals(Object other) {
		if(this == other){
			return true;
		}
		if(other == null){
			return false;
		}
		if(getClass() != other.getClass()){
			return false;
		}
		YMD a = (YMD)other;
		return (yy == a.yy) && (mm == a.mm) && (dd == a.dd);
	}

	@Override
	public int hashCode() {
		return 17
				* Integer.valueOf(yy).hashCode()
				+ 19
				* (Integer.valueOf(mm).hashCode() * 23 * (Double.valueOf(dd).hashCode()));
	}

	@Override
	public int compareTo(YMD ymd) {
		double a = dateToJulianDay(yy, mm, dd);
		double b = dateToJulianDay(ymd);
		return Double.compare(a, b);
	}

	@Override
	public String toString() {
		return yy + " " + mm + " " + dd;
	}

	//------------------------- static methods -------------------------------
	/**
	 * Podaje liczbe dni od 1st January 4713 B.C. czyli od południa
	 * Jan 1st 4713 B.C. mierzonego na południku Greenwich (czyli każdy nowy
	 * dzień juliański zaczyna sie o 12.00 UT (Universal Time)
	 * A. D (anno domini = roku panskiego)
	 * B. C (before Christ) - era chrzescijanska zaczyna sie od roku 1 A.D. Roku 0
	 * nie ma w ogole. Rok wczesniej to 1 B.C. Dla celów astronomicznych
	 * przyjmuje sie, że jest rok zerowy. Lata B.C. są liczone jako lata 
	 * ujemne dlatego np. faktyczny rok 10 B.C
	 * jest astronomicznym rokiem -9, a 4713 B.C jest -4712, czyli od faktycznego
	 * roku B. C. odejmuje sie zawsze jeden i zmienia znak na przeciwny. Jesli 
	 * rok astronomiczny jest ujemny zmienia sie znak na przeciwny, dodaje jeden
	 * i otrzymuje rok B.C. 
	 * Algorytm napisany na podstawie opisu algorytmu zawartego w książce:
	 * Jean Meeus. Astronomical Algorithms. Wyd. II
	 * @return - zwraca datę w postaci Julian day
	 */
	public static double dateToJulianDay(int rok, int miesiac, double dzien) {
		int y = rok;
		int m = miesiac;
		int a;
		int b = 0;
		if(m < 3){
			y--;
			m += 12;
		}
		if(isGCDate(rok, miesiac, dzien)){
			a = y / 100;
			b = 2 - a + (a / 4);
		}
		return (int)(365.25 * (y + 4716)) + (int)(30.6001 * (m + 1)) + dzien
				+ b - 1524.5;
	}

	/**
	 * Podaje liczbe dni od 1st January 4713 B.C. czyli od południa
	 * Jan 1st 4713 B.C. mierzonego na południku Greenwich (czyli każdy nowy
	 * dzień juliański zaczyna sie o 12.00 UT (Universal Time)
	 * A. D (anno domini = roku panskiego)
	 * B. C (before Christ) - era chrzescijanska zaczyna sie od roku 1 A.D. Roku 0
	 * nie ma w ogole. Rok wczesniej to 1 B.C. Dla celów astronomicznych
	 * przyjmuje sie, że jest rok zerowy. Lata B.C. są liczone jako lata 
	 * ujemne dlatego np. faktyczny rok 10 B.C
	 * jest astronomicznym rokiem -9, a 4713 B.C jest -4712, czyli od faktycznego
	 * roku B. C. odejmuje sie zawsze jeden i zmienia znak na przeciwny. Jesli 
	 * rok astronomiczny jest ujemny zmienia sie znak na przeciwny, dodaje jeden
	 * i otrzymuje rok B.C. 
	 * Algorytm napisany na podstawie opisu algorytmu zawartego w książce:
	 * Jean Meeus. Astronomical Algorithms. Wyd. II
     * @return - zwraca datę w postaci Julian Day
	 */
	public static double dateToJulianDay(YMD ymd) {
		int y = ymd.getYy();
		int m = ymd.getMm();
		double d = ymd.getDd();
		int a;
		int b = 0;
		if(m < 3){
			y--;
			m += 12;
		}
		if(isGCDate(ymd)){
			a = y / 100;
			b = 2 - a + (a / 4);
		}
		return (int)(365.25 * (y + 4716)) + (int)(30.6001 * (m + 1)) + d
				+ b - 1524.5;
	}

	/**
	 * Sprawdza czy podana data jest datą w Gregorian Calendar
	 * @return - zwraca true, jeśli podana data jest datą gregoriańską, a
	 * false, jeśli jest datą juliańską.
	 */
	public static boolean isGCDate(int rok, int miesiac, double dzien) {
		if(rok > 1582){
			return true;
		}
		if(rok < 1582){
			return false;
		}
		if(miesiac > 10){
			return true;
		}
		if(miesiac < 10){
			return false;
		}
		if(dzien >= 15.5){
			return true;
		}
		if(dzien <= 4.5){
			return false;
		}
		if(dzien > 4.5){
			throw new IllegalArgumentException("taka data nie istniała");
		}
		return false;
	}

	/**
	 * Sprawdza czy podana data jest datą w Gregorian Calendar
	 * @param ymd - data
	 * @return - zwraca true, jeśli podana data jest datą gregoriańską, a
	 * false, jeśli jest datą juliańską.
	 */
	public static boolean isGCDate(YMD ymd) {
		int rok = ymd.getYy();
		int miesiac = ymd.getMm();
		double dzien = ymd.getDd();
		if(rok > 1582){
			return true;
		}
		if(rok < 1582){
			return false;
		}
		if(miesiac > 10){
			return true;
		}
		if(miesiac < 10){
			return false;
		}
		if(dzien >= 15.5){
			return true;
		}
		if(dzien <= 4.5){
			return false;
		}
		if(dzien > 4.5){
			throw new IllegalArgumentException("taka data nie istniała");
		}
		return false;
	}

	/**
	 * Oblicza Julian Day dla początku podanego roku
	 * Algorytm napisany na podstawie opisu algorytmu zawartego w książce:
	 * Jean Meeus. Astronomical Algorithms. Wyd. II
	 * @return - zwraca  Julian Day dla January 0.0 tego roku.
	 */
	public static double julianDayStartOfYear(int rok) {
		int a;
		int y = rok;
		y--;		
                a = (y / 100);
		return (int)(365.25 * y) - a + (a / 4) + 1721424.5;
	}

	/**
	 * sprawdza czy podany rok jest przestępny
	 * @return - zwraca true, jeśli podany rok jest przestępny, false -
	 * w przeciwnym wypadku
	 */
	public static boolean isLeapYear(int rok) {
		if(rok == 1582){
			return false;
		}
		if(rok < 1582){
                    return (rok % 4) == 0;
		}
		if(!(String.valueOf(rok)).endsWith("00")){
return (rok % 4) == 0;
		}
		else{
return (rok % 400) == 0;
		}
	}

	/**
	 * Oblicza tzw. zmodyfikowany dzień juliański, czyli dzień juliański
	 * liczony od 17th Nov 1858, 0h (godz. 0.0) jako daty 0.
	 * Algorytm napisany na podstawie opisu algorytmu zawartego w książce:
	 * Jean Meeus. Astronomical Algorithms. Wyd. II
	 * @return - zwraca zmodyfikowany dzień juliański
	 */
	public static double modifiedJulianDay(int rok, int miesiac, double dzien) {
		double d = dateToJulianDay(rok, miesiac, dzien);
		d = d - 2400000.5;
		return d;
	}

	/**
	 * Oblicza datę dla podanego Julian day.
	 * @param julianDay - dzień juliański. Nie może być ujemny. Działa w
	 * zakresie dat kalendarza juliańskiego
	 * Algorytm napisany na podstawie opisu algorytmu zawartego w książce:
	 * Jean Meeus. Astronomical Algorithms. Wyd. II
	 * @return - zwraca obiekt daty
	 */
	public static YMD julianDayToDate(double julianDay) {
		if(julianDay < 0){
			throw new IllegalArgumentException(
					"Data nie może być wcześniejsza niż 1 stycznia 4713 B.C.");
		}
		int data0;
		int data1;
		double data2;
		julianDay += 0.5;
		int i = (int)julianDay;
		double fract = julianDay - i;
		int b;
		if(i >= 2299161){
			int A = (int)((i - 1867216.25) / 36524.25);
			b = i + 1 + A - (A / 4);
		}
		else{
			b = i;
		}
		int c = b + 1524;
		int d = (int)((c - 122.1) / 365.25);
		int e = (int)(365.25 * d);
		int g = (int)((c - e) / 30.6001);
		double dzien = c - e + fract - (int)(30.6001 * g);
		data2 = roundToDecimal(dzien, 4);
		int miesiac = -1;
		if(g < 13.5){
			miesiac = g - 1;
		}
		if(g > 13.5){
			miesiac = g - 13;
		}
		data1 = miesiac;
		int rok = 0;
		if(miesiac > 2.5){
			rok = d - 4716;
		}
		if(miesiac < 2.5){
			rok = d - 4715;
		}
		data0 = rok;
		return new YMD(data0, data1, data2);
	}

	/**
	 * Oblicza odstęp między datami (w dniach)
	 * @param r1 - rok 1
	 * @param m1 - miesiąc 1
	 * @param d1 - dzień 1
	 * @param r2 - rok 2
	 * @param m2 - miesiąc 2
	 * @param d2 - dzień 2
	 * @return - odstęp między datami. Ułamkowa częśc zawiera godziny, etc.
	 */
	public static double timeElapsed(int r1, int m1, double d1, int r2, int m2,
			double d2) {
		double jd1 = dateToJulianDay(r1, m1, d1);
		double jd2 = dateToJulianDay(r2, m2, d2);
		return Math.abs(jd2 - jd1);
	}

	/**
	 * Oblicza odstęp między datami.  
	 * @param date1 - data 1
	 * @param date2 - data 2
	 * @return - odstęp między datami. Ułamkowa część zawiera godziny
	 */
	public static double timeElapsed(YMD date1, YMD date2) {
		double jd1 = dateToJulianDay(date1);
		double jd2 = dateToJulianDay(date2);
		return Math.abs(jd2 - jd1);
	}

	/**
	 * Dodaje wskazana liczbę dni do podanej daty.
	 * @param r - rok
	 * @param m - miesiąc
	 * @param d - dzień
	 * @param days - liczba dni, które należ dodać
	 * @return - zwraca nową datę
	 */
	public static YMD addDays(int r, int m, double d, double days) {
		double jd = dateToJulianDay(r, m, d);
		jd += days;
		return julianDayToDate(jd);
	}

	/**
	 * Dodaje wskazana liczbę dni do podanej daty.
	 * @param ymd - data
	 * @param days - liczba dni, które należ dodać
	 * @return - zwraca nową datę po zmianach
	 */
	public static YMD addDays(YMD ymd, double days) {
		double jd = dateToJulianDay(ymd.getYy(), ymd.getMm(), ymd.getDd());
		jd += days;
		return julianDayToDate(jd);
	}

	/**
	 * Odejmuje wskazaną liczbę dni od podanej daty
	 * @param ymd - data
	 * @param days - liczba dni, które należy odjąć
	 * @return - zwraca nową datę po zmianach
	 */
	public static YMD substractDays(YMD ymd, double days) {
		double jd = dateToJulianDay(ymd.getYy(), ymd.getMm(), ymd.getDd());
		jd -= days;
		YMD result = julianDayToDate(jd);
		if(isValidDate(result)){
			return result;
		}
		return null;
	}

	/**
	 * Odejmuje wskazaną liczbę dni od podanej daty
	 * @param r - rok
	 * @param m - miesiąc
	 * @param d - dzień
	 * @param days - liczba dni, które należ odjąć
	 * @return - zwraca nową datę po zmianach
	 */
	public static YMD substractDays(int r, int m, double d, double days) {
		double jd = dateToJulianDay(r, m, d);
		jd -= days;
		YMD result = julianDayToDate(jd);
		if(isValidDate(result)){
			return result;
		}
		return null;
	}

	/**
	 * Podaje dzień tygodnia dla podanego Julian Day.
	 * Algorytm napisany na podstawie opisu algorytmu zawartego w książce:
	 * Jean Meeus. Astronomical Algorithms. Wyd. II
	 * @param julDay - dzień juliański
	 * @return - zwraca nr dnia tygodnia. 0 - niedziela, 1 - poniedziałek, etc
	 */
	public static int dayOfWeek(double julDay) {
		double A = (julDay + 1.5) / 7;
		return (int)((A - (int)A) * 7);
	}

	/**
	 * Podaje dzień tygodnia dla podanej daty.	 * 
	 * @param r - rok
	 * @param m - miesiąc
	 * @param d - dzień
	 * @return - zwraca nr dnia tygodnia. 0 - niedziela, 1 - poniedziałek, etc
	 */
	public static int dayOfWeek(int r, int m, double d) {
		double julDay = dateToJulianDay(r, m, d);
		return dayOfWeek(julDay);
	}

	/**
	 * Podaje dzień tygodnia dla podanej daty.
	 * @param ymd - data
	 * @return - zwraca nr dnia tygodnia. 0 - niedziela, 1 - poniedziałek, etc
	 */
	public static int dayOfWeek(YMD ymd) {
		double julDay = dateToJulianDay(ymd);
		return dayOfWeek(julDay);
	}

	/**
	 * Oblicza nr kolejnego dnia w roku dla podanej daty.
	 * Algorytm napisany na podstawie opisu algorytmu zawartego w książce:
	 * Jean Meeus. Astronomical Algorithms. Wyd. II
	 * @param r - rok
	 * @param m - miesiąc
	 * @param d - dzień
	 * @return - zwraca nr kolejny dnia w roku dla podanej daty
	 */
	public static int dayOfYear(int r, int m, double d) {
		int k;
		if(!isLeapYear(r)){
			k = 2;
		}
		else{
			k = 1;
		}
		return ((275 * m) / 9) - k * (((m + 9) / 12)) + (int)d - 30;
	}

	/**
	 * Oblicza nr kolejnego dnia w roku dla podanej daty.
	 * Algorytm napisany na podstawie opisu algorytmu zawartego w książce:
	 * Jean Meeus. Astronomical Algorithms. Wyd. II
	 * @param ymd - data
	 * @return - zwraca nr kolejny dnia w roku dla podanej daty
	 */
	public static int dayOfYear(YMD ymd) {
		int k;
		if(!isLeapYear(ymd.getYy())){
			k = 2;
		}
		else{
			k = 1;
		}
		return ((275 * ymd.getMm()) / 9) - k * (((ymd.getMm() + 9) / 12))
				+ (int)ymd.getDd() - 30;
	}

	/**
	 * Sprawdza czy podana data jest poprawną datą
	 * @param year - rok
	 * @param month - miesiąc
	 * @param day - dzień
	 * @return - zwraca true, jeśli data jest poprawna, false, jeśli data
	 * nie jest poprawna
	 */
	public static boolean isValidDate(int year, int month, double day) {
		if((month >= 1) && (month <= 12)){
			if(dateToJulianDay(year, month, day) >= 0.0){
				if(!isLeapYear(year)){
					return ((int) day >= 1) && ((int) day <= LEN[month - 1]);
				}
				else{
					return ((int) day >= 1) && ((int) day <= LEAP_LEN[month - 1]);
				}
			}
			else{
				throw new IllegalArgumentException(
						"data poza zakresem kalendarza");
			}
		}
		return false;
	}

	/**
	 * Sprawdza czy podana data jest poprawną datą
	 * @param ymd  - data
	 * @return - zwraca true, jeśli data jest poprawna, false, jeśli data
	 * nie jest poprawna
	 */
	public static boolean isValidDate(YMD ymd) {
		if((ymd.getMm() >= 1) && (ymd.getMm() <= 12)){
			if(dateToJulianDay(ymd.getYy(), ymd.getMm(), ymd.getDd()) >= 0.0){
				if(!isLeapYear(ymd.getYy())){
					return ((int) ymd.getDd() >= 1)
							&& ((int) ymd.getDd() <= LEN[ymd.getMm() - 1]);
				}
				else{
					return ((int) ymd.getDd() >= 1)
							&& ((int) ymd.getDd() <= LEAP_LEN[ymd.getMm() - 1]);
				}
			}
			else{
				throw new IllegalArgumentException(
						"data poza zakresem kalendarza");
			}
		}
		return false;
	}

	/**
	 * Podaje datę zmiany czasu letniego na zimowy w podanym roku
	 * @param year - rok
	 * @return - zwraca datę zmiany czasu letniego na zimowy w podanym roku
	 */
	public static YMD dateOfSummerToWinter(int year) {
		YMD ymd = new YMD(year, 10, 31.5);
		int day = dayOfWeek(ymd);
		ymd.substractDays(day);
		return ymd;
	}

	/**
	 * Podaje datę zmiany czasu zimowego na letni w podanym roku
	 * @param year - rok
	 * @return - zwraca datę zmiany czasu zimowego na letni w podanym roku
	 */
	public static YMD dateOfWinterToSummer(int year) {
		YMD ymd = new YMD(year, 3, 31.5);
		int day = dayOfWeek(ymd);
		ymd.substractDays(day);
		return ymd;
	}

	/** 
	 * Metoda zaokrągla podaną liczbę do określonej liczby miejsc po przecinku
	 *@param num - liczba do zaokrąglenia
	 *@param dec  - liczba miejsc po przecinku;
	 *@return double - zwraca zaokrągloną liczbę
	 */
	public static double roundToDecimal(double num, int dec) {
		int multi = (int)Math.pow(10, dec);
		int temp = (int)Math.round(num * multi);
		return (double)temp / multi;
	}
}
